/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      Audio input test program for the Allegro library.
 *
 *      By Shawn Hargreaves.
 *
 *      See readme.txt for copyright information.
 */


#include <stdio.h>
#include <string.h>
#include <math.h>

#include "allegro.h"

BITMAP *scbuf, *imbuf;
BITMAP *lettbuf;

//#define TUNERFRAGSIZE 8192
#define TUNERFRAGSIZE 4096
//#define TUNERFRAGSIZE 2048

// sinx/x interpol params
#define IFACT 8
#define INTERP_LEN 16


double *a, *w, *calib, *falloff;
double interpolate_buffer[TUNERFRAGSIZE * IFACT];
int *ip;
int octavelen , octavestart, num_octaves, descend;
    int count = 0;

double foundfreq;

extern void rdft(int n, int isgn, double *a, int *ip, double *w);
void interpolate_sinx_x(double *inbuf, int inlen, double *outbuf );

// dirty bitmap hack
void rotatebitmap(BITMAP *bmp, int amount)
{
    static char **l = NULL;
    static int maxh = 0;
    int i;

    while(amount < 0)
        amount += bmp->h;
    
    if (bmp->h > maxh)
    {
        if (l)
            free(l);

        l = (char **)malloc(sizeof(char *) * bmp->h);
        maxh = bmp->h;

        ASSERT(l);
    }
    
    
    for (i = 0; i < bmp->h; i++)
    {
        l[i] = bmp->line[(i + amount) % bmp->h];
    }

    for (i = 0; i < bmp->h; i++)
    {
        bmp->line[i] = l[i];
    }


    
}


void init_fft()
{
    int i;
    a = malloc(sizeof(double) * TUNERFRAGSIZE);
    calib = malloc(sizeof(double) * TUNERFRAGSIZE * IFACT/2);
    falloff = malloc(sizeof(double) * TUNERFRAGSIZE * IFACT/2);
    w = malloc(sizeof(double) * (TUNERFRAGSIZE*IFACT/2));
    ip = malloc(sizeof(int) * (2 + sqrt(TUNERFRAGSIZE *IFACT/ 2)));

    for (i = 0; i< TUNERFRAGSIZE/2; i++)
        falloff[i] = calib[i] = 0;
    
    ASSERT(a);
    ASSERT(w);
    ASSERT(ip);                   
    ASSERT(falloff);
    ASSERT(calib);
    
    ip[0] = 0;
    
    octavelen = octavestart = 2*880 *TUNERFRAGSIZE/ 44100;
    num_octaves = (int)(log(TUNERFRAGSIZE / 110)/ log(2));
    descend = 3;
    
}
                  

void process_frag()
{   
    int i,j;
    double ac;


    interpolate_sinx_x(a, TUNERFRAGSIZE, interpolate_buffer);
    
    rdft(TUNERFRAGSIZE * IFACT, -1, interpolate_buffer, ip, w);
    
    for (i = 0; i < TUNERFRAGSIZE/2;i++)
    {                                                                            
        interpolate_buffer[i] = interpolate_buffer[2*i] * interpolate_buffer[2*i] + interpolate_buffer[2*i+1] * interpolate_buffer[2*i+1];

        ac = interpolate_buffer[i] - calib[i];


        if (count < 100)
            calib[i] = (3 * calib[i] + interpolate_buffer[i]) / 4;

        if (ac < 0)
            ac = 0;
        
        falloff[i] =  2* falloff[i] + ac;
        falloff[i] /= 3;
        
    }
    
}

void findpeak()
{
    double max = 0;
    int peak = -1;
    int i;
    int from, to;
    int fpeak = -1;
    double delta;

    static double tmpbuf[(4*INTERP_LEN + 2) * IFACT];
    for (i = 50 * TUNERFRAGSIZE*IFACT/ 44100 ; i < TUNERFRAGSIZE*IFACT/4;i++)
    {
        if (falloff[i] > max)
        {
            max = falloff[i];
            peak = i;
        }
    }

    if (peak < 0)
        return;

    foundfreq = peak;


    foundfreq *= 44100;
    foundfreq /= TUNERFRAGSIZE;

}


int hcolor(double f)
{
    static double min = 1000000000;
    static double max = -1000000000;
    int c;


    if (key[KEY_SPACE])
    {
        min = 1000000000;
        max = -1000000000;
    }

    if (f < min)
        min = f;

    if (f/10 > max)
        max = f/10;

    min = 0;
    max = 400;

    c = (int)((f - min) * 255*4 / (max - min));

    c = MIN(1023,c);
    c = MAX(0,c);

//    printf("%g %g %g %d\n", min, max, f,c);

    if (c < 256)
        return makecol(c , 0,255);
    else if (c < 512)
        return makecol(255, 0, 511 - c);
    else if (c < 768)
        return makecol(255, c-512, 0);

    return makecol(255, 255, c - 768);
}

void draw()
{
    int i;
    static int count = 0;
    int dx, dy;
    double angle;
    static double cangle = 0;
    double dist;
    int octave;
    int note;
    char *notename[13] = {"A", "A#", "B","C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A"};
    
    
//clear_to_color(scbuf, bitmap_mask_color(scbuf));
    clear_to_color(lettbuf, makecol(0,0,0));

  /*
    vline(scbuf, 440 *TUNERFRAGSIZE/44100,0, 480, makecol(0,0,255)) ;
    vline(scbuf, 880 *TUNERFRAGSIZE/44100,0, 480, makecol(0,0,255)) ;
    vline(scbuf, 2*880 *TUNERFRAGSIZE/44100,0, 480, makecol(0,0,255)) ;

    rotatebitmap(imbuf, -1);

    for (i = 0; i< TUNERFRAGSIZE / 2; i++)
    {
        vline(scbuf, i, 0, falloff[i], makecol(255,0,0));
        putpixel(imbuf, i,0, hcolor(falloff[i]));
    }



    */




    textprintf_centre(lettbuf, font, lettbuf->w/2,1,makecol(55,55,55), "FREQ %g ", foundfreq);

    octave = 0;
    while (foundfreq > 112)
    {
        foundfreq/=2;
        octave +=1;
    }
    if (foundfreq < 55.0)
        foundfreq = 40.0;

    angle = (log(foundfreq) - log(55.0))/(log(110.0) - log(55.0));

    note = (int)(angle * 12.0 + .5);

    while (note > 12)
    {
        note-=12;
        octave +=1;
    }
    while (note < 0)
    {
        octave -=1;
        note +=12;
    }
    
    textprintf_centre(lettbuf, font,lettbuf->w/2, 10,makecol(55,55,55), "octave %d", octave);
    textprintf_centre(lettbuf, font,lettbuf->w/2, 20,makecol(55,55,55), "note %s", notename[note]);

    stretch_blit(lettbuf, scbuf, 0,0, lettbuf->w, lettbuf->h, 0,0, SCREEN_W, SCREEN_H);

    for (i = 0; i< 13; i++)
    {
        dx = SCREEN_W/2* cos(M_PI * ((i / 12.0) * .8 + .1));
        dy = SCREEN_W/2* sin(M_PI * ((i / 12.0) * .8 + .1));
        line(scbuf, SCREEN_W/2-dx/2, SCREEN_H-dy/2-1, SCREEN_W/2  -dx, SCREEN_H - dy-1, makecol(100,155,0));
    }
    

    angle *=.8;
    angle += .1;

    dist = angle - cangle;

    cangle = cangle + dist/13.0;



    dx = SCREEN_W/2* cos(M_PI * cangle);
    dy = SCREEN_W/2* sin(M_PI * cangle);

    line(scbuf, SCREEN_W/2, SCREEN_H, SCREEN_W/2  -dx, SCREEN_H - dy, makecol(0,255,0));


//  blit(imbuf,screen,0,0,0,0,SCREEN_W,SCREEN_H);
// masked_blit(scbuf,screen,0,0,0,0,SCREEN_W,SCREEN_H);
    blit(scbuf,screen,0,0,0,0,SCREEN_W,SCREEN_H);

    count++;
}


double sinx_xtab[IFACT * (INTERP_LEN + 1)];

void init_sinx()
{
    int i;
    sinx_xtab[0] = 1.0;
    for (i = 1; i< IFACT * (INTERP_LEN+1); i++)
    {

        sinx_xtab[i] = sin(M_PI * i/ IFACT) * IFACT / (i * M_PI);
    }
}

double sinx_x(int angle)
{
    if (angle < 0)
        angle = -angle;

    if (angle >= IFACT * (INTERP_LEN + 1))
    {
        printf("warning , angle > %d PI radians, slow sin (x)/x used\n", INTERP_LEN);
        return sin(M_PI * angle/ IFACT) * IFACT/ angle;
    }
    return sinx_xtab[angle];
}


void interpolate_sinx_x(double *inbuf, int inlen, double *outbuf )
{
    int i, j;
    int outlen = inlen * IFACT;
    int pos,d;
    
    for (i = 0; i< outlen; i++)
    {
        outbuf[i] = 0;
        pos = i / IFACT;
        d = -(i % IFACT);
        for (j = -INTERP_LEN; j < INTERP_LEN; j++)
        {
            if (pos + j >=0 && pos +j < inlen)
                outbuf[i] += inbuf[pos + j] * sinx_x(j * IFACT + d );
        }
    }
}



double fn(int i)
{
    double x =  i /5.0 - 4.0;
    return -exp(-x*x);
}

void test_interpolation()
{

    double in[164], out[164 *IFACT];
    int i;
    char buffer[256];
    int count;

    FILE *inf = fopen("dump.txt", "r");

    fgets(buffer, 256, inf);


    count = atoi(buffer);

    for (i ; i< count ; i++)
    {
        fgets(buffer, 256, inf);
        in[i] = strtod(buffer, NULL) / 200;
        printf("%g\n", in[i]);
    }

    fclose(inf);



    interpolate_sinx_x(in, count, out);

    for (i = 0; i < count*IFACT; i++)
    {
//        putpixel(screen, i, SCREEN_W/2 + fn(i), makecol(255,0,0));
        if (!(i % IFACT))
            circlefill(screen, i, SCREEN_W/2 + (int)(in[i/IFACT]), 1, makecol(0,0,255));
        putpixel(screen, i, SCREEN_W/2 + (int)(out[i]), makecol(0,255,0));




    }


    readkey();

}


void tuner()
{
    int bufsize;
    unsigned short *buffer;    
    int i;
    int bufpos = 0;
    
    bufsize = start_sound_input(44100, 16, 1);
    bufsize/= 2;
    
    
    if (bufsize <= 0) {
       alert("Error recording", NULL, NULL, "Drat", NULL, 13, 0);
       return;
    }         
                                                      
    ASSERT(sizeof(unsigned short) == 2);
    
    buffer = malloc(sizeof(unsigned short) * bufsize);
                                   
    ASSERT(buffer);

    init_fft();
     
    while(!key[KEY_ESC])
    {
        if (read_sound_input(buffer))
        {          
            for (i = 0; i< bufsize/2; i++)
            {
                if (i+ bufpos < TUNERFRAGSIZE)
                    a[i + bufpos] = (buffer[2*i] + buffer[2*i + 1] )/ 131072.0;
            }
            bufpos += bufsize/2;
            
            if (bufpos >= TUNERFRAGSIZE)
            {
                count++;
                process_frag();
                findpeak();
                    
                draw();
                bufpos = 0;
            }
        }   
    }
    
    
    stop_sound_input();
    
}





int main()
{
   allegro_init();
   install_keyboard();
   install_mouse();
   install_timer();

   if (install_sound(DIGI_OSS, MIDI_AUTODETECT, NULL) != 0) {
      allegro_message("\nError initialising sound\n%s\n", allegro_error);
      return 1;
   }

   if (install_sound_input(DIGI_OSS, MIDI_NONE) != 0) {
      allegro_message("\nError initialising sound input\n%s\n\n"
            "Try changing the digi_input_card and midi_input_card settings in your\n"
            "allegro.cfg, in particular disabling (set to zero) one of those drivers.\n",
            allegro_error);
      return 1;
   }

   set_color_depth(16);
   if (set_gfx_mode(GFX_AUTODETECT, 800, 600,0, 0) != 0) {
      set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
      allegro_message("Error setting graphics mode\n%s\n", allegro_error);
      return 1;
   }


   set_palette(desktop_palette);
 
   scbuf = create_bitmap(SCREEN_W, SCREEN_H);
   imbuf = create_bitmap(SCREEN_W, SCREEN_H);
   lettbuf = create_bitmap(200, 30);
   
   ASSERT(scbuf);
   ASSERT(imbuf);
   ASSERT(lettbuf);

   init_sinx();
//   test_interpolation();

   
   tuner();




   return 0;
}

END_OF_MAIN();





















